/* This file is part of the 'neper' program. */
/* Copyright (C) 2004-2009, Romain Quey. */
/* See the COPYING file in the top-level directory. */

#include "SpaceUtils.h"

/* threeDDoubleDist returns the euclidian distance between two points.*/
double
twoDDoubleDist (double *a, double *b, int beg)
{
  double res;

  if (beg == 1)
    res = sqrt ((b[1] - a[1])*(b[1] - a[1]) + (b[2] - a[2])*(b[2] - a[2]));
  else if (beg == 0)
    res = sqrt ((b[0] - a[0])*(b[0] - a[0]) + (b[1] - a[1])*(b[1] - a[1]));
  else
    abort ();

  return res;
}

/* threeDDoubleDist returns the euclidian distance between two points.*/
double
threeDDoubleDist (double *a, double *b, int beg)
{
  double res;

  if (beg == 1)
    res =
      sqrt (pow (b[1] - a[1], 2) + pow (b[2] - a[2], 2) +
	    pow (b[3] - a[3], 2));
  else if (beg == 0)
    res =
      sqrt (pow (b[0] - a[0], 2) + pow (b[1] - a[1], 2) +
	    pow (b[2] - a[2], 2));
  else
    abort ();

  return res;
}


/* Side indicates the half-space whose a specified point belongs to. */
int
Side (double *P, double *C)
{
  if (P[1] * C[1] + P[2] * C[2] + P[3] * C[3] - P[0] < 0)
    return -1;
  else
    return 1;
}

/* Side indicates the half-space whose a specified point belongs to. */
int
Side_tol (double *P, double *C, double eps)
{
  if (P[1] * C[1] + P[2] * C[2] + P[3] * C[3] - P[0] < eps)
    return -1;
  else
    return 1;
}

/* Side2d indicates the half-space whose a specified point belongs to (2d). */
int
Side2d (double *P, double *C)
{
  if (P[1] * C[1] + P[2] * C[2] - P[0] < 0)
    return -1;
  else
    return 1;
}


/* Barycenter calculates the coordinates of the polygon/hedron
 * center, as the barycenter of all its vertices.
 */
void
Barycenter (double **coo, int Vqty, double *res)
{
  int i, j;
  double tmp;

  for (i = 1; i <= 3; i++)
    {
      tmp = 0;
      for (j = 1; j <= Vqty; j++)
	tmp += coo[j][i];

      res[i] = tmp / (double) Vqty;
    }

  return;
}

/* The volume of a tetrahedron is 1/3 . A.H, where A is the area of
 * one of the faces, H the corresponding tetrahedron height.
 * Below, coo contains the coordinates of the 4 tetrahedron vertices.
 */
double
TetrahedronVolume (double *coo1, double *coo2, double *coo3, double *coo4)
{
  int i;
  double A, H, V;		/* area, height, volume */
  double * PlaneEq = oneDDoubleDAlloc (4);
  double * proj = oneDDoubleDAlloc (3);

  /* A is calculated for the triangle build from the 3 first vertices. */
  A = TriangleArea (coo1, coo2, coo3);

  /* calculating the equation of the previous triangle */
  Points2Plane (coo1, coo2, coo3, PlaneEq);

  /* calculating proj[] as the projection of the 4th ver onto the
   * previous triangle.
   */
  for (i = 0; i <= 2; i++)
    proj[i] = coo4[i];
  ProjectPointAlongOnto (proj, PlaneEq + 1, PlaneEq);

  H = threeDDoubleDist (coo4, proj, 0);

  V = A * H / 3;

  oneDDoubleFree (PlaneEq);
  oneDDoubleFree (proj);

  return V;
}

/* PolygonArea returns the area of a polygon.
 * THE POLYGON MUST BE CONVEX for PolygonArea to return a reliable
 * result. The adopted algorithm consists in the decomposition of
 * the polygon into as many triangles as there are polygon vertices
 * or edges, and to sum their individual contributions. A central
 * point is used to build these triangles.
 * vertex i: coo[i][1,2,3] (i=1,2,3)
 * N.B.: for triangles, the specific subroutine TriangleArea
 * can also be used.
 */
/*
double PolygonArea(struct POLYGON Polygon)
{
    int i;	
    double res=0;
    double** coo;
    
    coo=(double**)twoDDoubleDAlloc(4,3);
    
    for(i=1;i<=Polygon.VerQty-1;i++)
    {
	PolygonTriangle(Polygon,i,i+1,coo);
	res+=TriangleArea(coo);
    }
    PolygonTriangle(Polygon,Polygon.VerQty,1,coo);
    res+=TriangleArea(coo);

    twoDDoubleFree(coo,4);
    
    return res;
}
*/

/* PolyhedronVolume returns the area of a polygon. THE POLYHEDRON
 * MUST BE CONVEX for PolyhedronVolume to return a reliable
 * result. The adopted algorithm consists in the decomposition of
 * the polyhedron into as many pyramids as there are polyhedron faces
 * and to sum their individual contributions. A central
 * point is used to build these pyramids.
 */
/*
double PolyhedronVolume(struct POLY Poly)
{
    double res=0;
    int i,k;
    double*  PCenter;  
    double** TrCoo;
    double** coo;

    TrCoo=  (double**)twoDDoubleDAlloc( 4,4);
    PCenter=(double*) oneDDoubleDAlloc( 4);


    for(i=1;i<=Poly.VerQty;i++)
    {
	printf("i=%d/%d\n",i,Poly.VerQty);
	for(k=0;k<=2;k++)
	{
	    printf("k=%d\n",k);
	    printf("%lf\n",Poly.VerCoo[i][k]);
	    coo[i][k+1]=Poly.VerCoo[i][k];
	}
    }

    printf("hi1\n");
    Barycenter(coo,Poly.VerQty,PCenter);
    printf("hi2\n");

    return res;
}
*/

void
TriangleN (double *p1, double *p2, double *p3, double *n, int beg)
{
  int i;
  double * a = oneDDoubleDAlloc (4);
  double * b = oneDDoubleDAlloc (4);
  double norm;

  if (beg != 0 && beg != 1)
    abort ();

  for (i = 0; i <= 2; i++)
    {
      a[i + beg] = p2[i + beg] - p1[i + beg];
      b[i + beg] = p3[i + beg] - p2[i + beg];
    }
  VectorVectProd (a, b, n, beg);
  norm = VectorNorm (n, beg);

  for (i = 0; i <= 2; i++)
    n[i + beg] /= norm;

  oneDDoubleFree (a);
  oneDDoubleFree (b);

  return;
}

/* Points2Line gives the equation of a line from the coo of 2 points. */
void
Points2Line (double *p1, double *p2, double *eq)
{
  double norm;

  int s;
  gsl_matrix *A = gsl_matrix_alloc (3, 3);
  gsl_vector *B = gsl_vector_alloc (3);
  gsl_vector *X = gsl_vector_alloc (3);
  gsl_permutation *p = gsl_permutation_alloc (3);

  gsl_matrix_set (A, 0, 0, p1[0]);
  gsl_matrix_set (A, 0, 1, p1[1]);
  gsl_matrix_set (A, 0, 2, -1);
  gsl_vector_set (B, 0, 0);
  gsl_matrix_set (A, 1, 0, p2[0]);
  gsl_matrix_set (A, 1, 1, p2[1]);
  gsl_matrix_set (A, 1, 2, -1);
  gsl_vector_set (B, 1, 0);

  gsl_matrix_set (A, 2, 0, 1);
  gsl_matrix_set (A, 2, 1, 1);
  gsl_vector_set (B, 2, 1);

  gsl_linalg_LU_decomp (A, p, &s);
  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
    {
      gsl_matrix_set (A, 2, 0, 1);
      gsl_matrix_set (A, 2, 1, 0);
      gsl_linalg_LU_decomp (A, p, &s);
      if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	{
	  gsl_matrix_set (A, 2, 0, 0);
	  gsl_matrix_set (A, 2, 1, 1);
	  gsl_linalg_LU_decomp (A, p, &s);
	  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	    {
	      printf ("routine Points2Line\n");
	      abort ();
	    }
	}
    }
  gsl_linalg_LU_solve (A, p, B, X);

  norm = sqrt (pow (gsl_vector_get (X, 0), 2) +
	       pow (gsl_vector_get (X, 1), 2));

  gsl_vector_scale (X, 1 / norm);

  eq[0] = gsl_vector_get (X, 2);
  eq[1] = gsl_vector_get (X, 0);
  eq[2] = gsl_vector_get (X, 1);

  gsl_matrix_free (A);
  gsl_vector_free (B);
  gsl_vector_free (X);
  gsl_permutation_free (p);

  return;
}

/* Points2Plane gives the equation of a plane from the coo of 3 points. */
void
Points2Plane (double *p1, double *p2, double *p3, double *eq)
{
  double norm;

  int s;
  gsl_matrix *A = gsl_matrix_alloc (4, 4);
  gsl_vector *B = gsl_vector_alloc (4);
  gsl_vector *X = gsl_vector_alloc (4);
  gsl_permutation *p = gsl_permutation_alloc (4);

  gsl_matrix_set (A, 0, 0, p1[0]);
  gsl_matrix_set (A, 0, 1, p1[1]);
  gsl_matrix_set (A, 0, 2, p1[2]);
  gsl_matrix_set (A, 0, 3, -1);
  gsl_vector_set (B, 0, 0);
  gsl_matrix_set (A, 1, 0, p2[0]);
  gsl_matrix_set (A, 1, 1, p2[1]);
  gsl_matrix_set (A, 1, 2, p2[2]);
  gsl_matrix_set (A, 1, 3, -1);
  gsl_vector_set (B, 1, 0);
  gsl_matrix_set (A, 2, 0, p3[0]);
  gsl_matrix_set (A, 2, 1, p3[1]);
  gsl_matrix_set (A, 2, 2, p3[2]);
  gsl_matrix_set (A, 2, 3, -1);
  gsl_vector_set (B, 2, 0);

  gsl_matrix_set (A, 3, 0, 1);
  gsl_matrix_set (A, 3, 1, 1);
  gsl_matrix_set (A, 3, 2, 1);
  gsl_matrix_set (A, 3, 3, 0);
  gsl_vector_set (B, 3, 1);

  gsl_linalg_LU_decomp (A, p, &s);
  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
    {
      gsl_matrix_set (A, 3, 0, 1);
      gsl_matrix_set (A, 3, 1, 0);
      gsl_matrix_set (A, 3, 2, 0);
      gsl_linalg_LU_decomp (A, p, &s);
      if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	{
	  gsl_matrix_set (A, 3, 0, 0);
	  gsl_matrix_set (A, 3, 1, 1);
	  gsl_matrix_set (A, 3, 2, 0);
	  gsl_linalg_LU_decomp (A, p, &s);
	  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	    {
	      gsl_matrix_set (A, 3, 0, 0);
	      gsl_matrix_set (A, 3, 1, 0);
	      gsl_matrix_set (A, 3, 2, 1);
	      gsl_linalg_LU_decomp (A, p, &s);
	      if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
		{
		  printf ("routine Points2Plane\n");
		  abort ();
		}
	    }
	}
    }
  gsl_linalg_LU_solve (A, p, B, X);

  norm = sqrt (pow (gsl_vector_get (X, 0), 2) +
	       pow (gsl_vector_get (X, 1), 2) +
	       pow (gsl_vector_get (X, 2), 2));

  gsl_vector_scale (X, 1 / norm);

  eq[0] = gsl_vector_get (X, 3);
  eq[1] = gsl_vector_get (X, 0);
  eq[2] = gsl_vector_get (X, 1);
  eq[3] = gsl_vector_get (X, 2);

  gsl_matrix_free (A);
  gsl_vector_free (B);
  gsl_vector_free (X);
  gsl_permutation_free (p);

  return;
}

void
PointsNInVect2Plane (double *p1, double *p2, double *invect, double *eq)
{
  /* int i, j; */
  double norm;

  int s;
  gsl_matrix *A = gsl_matrix_alloc (4, 4);
  gsl_vector *B = gsl_vector_alloc (4);
  gsl_vector *X = gsl_vector_alloc (4);
  gsl_permutation *p = gsl_permutation_alloc (4);

  gsl_matrix_set (A, 0, 0, invect[0]);
  gsl_matrix_set (A, 0, 1, invect[1]);
  gsl_matrix_set (A, 0, 2, invect[2]);
  gsl_matrix_set (A, 0, 3, 0);
  gsl_vector_set (B, 0, 0);

  gsl_matrix_set (A, 1, 0, p1[0]);
  gsl_matrix_set (A, 1, 1, p1[1]);
  gsl_matrix_set (A, 1, 2, p1[2]);
  gsl_matrix_set (A, 1, 3, -1);
  gsl_vector_set (B, 1, 0);

  gsl_matrix_set (A, 2, 0, p2[0]);
  gsl_matrix_set (A, 2, 1, p2[1]);
  gsl_matrix_set (A, 2, 2, p2[2]);
  gsl_matrix_set (A, 2, 3, -1);
  gsl_vector_set (B, 2, 0);

  gsl_matrix_set (A, 3, 0, 1);
  gsl_matrix_set (A, 3, 1, 1);
  gsl_matrix_set (A, 3, 2, 1);
  gsl_matrix_set (A, 3, 3, 0);
  gsl_vector_set (B, 3, 1);

  /*
  printf ("beginning --------------------------------\n");
  printf ("A:\n");
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
      printf ("%f ", gsl_matrix_get (A, i, j));
    printf ("\n");
  }
	  
  printf ("B:\n");
  for (i = 0; i < 4; i++)
    printf ("%f\n", gsl_vector_get (B, i));
	  
	  
  printf ("X:\n");
  for (i = 0; i < 4; i++)
    printf ("%f\n", gsl_vector_get (X, i));
  printf ("end beginning ----------------------------\n");
  */

  gsl_linalg_LU_decomp (A, p, &s);

  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
    {
      gsl_matrix_set (A, 3, 0, 1);
      gsl_matrix_set (A, 3, 1, 0);
      gsl_matrix_set (A, 3, 2, 0);
      gsl_linalg_LU_decomp (A, p, &s);
      if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	{
	  gsl_matrix_set (A, 3, 0, 0);
	  gsl_matrix_set (A, 3, 1, 1);
	  gsl_matrix_set (A, 3, 2, 0);
	  gsl_linalg_LU_decomp (A, p, &s);
	  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
	    {
	      gsl_matrix_set (A, 3, 0, 0);
	      gsl_matrix_set (A, 3, 1, 0);
	      gsl_matrix_set (A, 3, 2, 1);
	      gsl_linalg_LU_decomp (A, p, &s);
	      if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
		{
		  printf ("routine PointsNInVect2Plane\n");
		  abort ();
		}
	    }
	}
    }

  gsl_linalg_LU_solve (A, p, B, X);
  
  /*
  printf ("A:\n");
  for (i = 0; i < 4; i++)
  {
    for (j = 0; j < 4; j++)
      printf ("%f ", gsl_matrix_get (A, i, j));
    printf ("\n");
  }
	  
  printf ("B:\n");
  for (i = 0; i < 4; i++)
    printf ("%f\n", gsl_vector_get (B, i));
	  
	  
  printf ("X:\n");
  for (i = 0; i < 4; i++)
    printf ("%f\n", gsl_vector_get (X, i));
  */

  norm = sqrt (pow (gsl_vector_get (X, 0), 2) +
	       pow (gsl_vector_get (X, 1), 2) +
	       pow (gsl_vector_get (X, 2), 2));

  gsl_vector_scale (X, 1 / norm);

  eq[0] = gsl_vector_get (X, 3);
  eq[1] = gsl_vector_get (X, 0);
  eq[2] = gsl_vector_get (X, 1);
  eq[3] = gsl_vector_get (X, 2);

  gsl_matrix_free (A);
  gsl_vector_free (B);
  gsl_vector_free (X);
  gsl_permutation_free (p);

  return;
}

/* Projection of a point along a vector onto a plane. */
void
ProjectPointAlongOnto (double *coo, double *n, double *plane)
{
  int i;
  int s;
  gsl_matrix *A = gsl_matrix_alloc (4, 4);
  gsl_vector *B = gsl_vector_alloc (4);
  gsl_vector *X = gsl_vector_alloc (4);
  gsl_permutation *p = gsl_permutation_alloc (4);

  /* Projection of any point M (X,Y,Z) along n(a',b',c') onto
   * plane P (ax+by+cz-d=0), denoted as M' (X',Y',Z') is given
   * by the following matrix system:
   * A.X=B with
   * ~ - -
   * A = | 1 0 0 a' |; X = |X'|; B = |X|. (k is a distance parameter.)
   *     | 0 1 0 b' |      |Y'|      |Y|  the 3 first lines means M'M
   *     | 0 0 1 c' |      |Z'|      |Z|  the 4th one means M' belongs to P.
   *     | a b c 0  |      |k |      |d|  

   * recording A, which does no depend upon ver, but which is modified by
   * gsl_linalg_LU_... . */
  gsl_matrix_set (A, 0, 0, 1);
  gsl_matrix_set (A, 0, 1, 0);
  gsl_matrix_set (A, 0, 2, 0);
  gsl_matrix_set (A, 0, 3, n[0]);
  gsl_vector_set (B, 0, coo[0]);

  gsl_matrix_set (A, 1, 0, 0);
  gsl_matrix_set (A, 1, 1, 1);
  gsl_matrix_set (A, 1, 2, 0);
  gsl_matrix_set (A, 1, 3, n[1]);
  gsl_vector_set (B, 1, coo[1]);

  gsl_matrix_set (A, 2, 0, 0);
  gsl_matrix_set (A, 2, 1, 0);
  gsl_matrix_set (A, 2, 2, 1);
  gsl_matrix_set (A, 2, 3, n[2]);
  gsl_vector_set (B, 2, coo[2]);

  gsl_matrix_set (A, 3, 3, 0);
  gsl_matrix_set (A, 3, 0, plane[1]);
  gsl_matrix_set (A, 3, 1, plane[2]);
  gsl_matrix_set (A, 3, 2, plane[3]);
  gsl_vector_set (B, 3, plane[0]);

  gsl_linalg_LU_decomp (A, p, &s);
  if (fabs (gsl_linalg_LU_det (A, s)) < 1e-15)
    {
      printf ("det=%.20f\n", gsl_linalg_LU_det (A, s));
      printf ("ProjectPointAlongOnto: det=0\n");
      abort ();
    }

  gsl_linalg_LU_solve (A, p, B, X);

  for (i = 0; i <= 2; i++)
    coo[i] = gsl_vector_get (X, i);

  gsl_matrix_free (A);
  gsl_vector_free (B);
  gsl_vector_free (X);
  gsl_permutation_free (p);

  return;
}

void
DistribPoints (double *MaxSize, double *MinSize, int Qty, double **Data,
	       gsl_rng * r)
{
  int i, j, id;
  double * coo = oneDDoubleDAlloc (4);

  /* ckecking the zone sizes */

  /* printf("DISTRIB from %lf to %lf: %d germs\n",-MinSize[1],-MaxSize[1],Qty); */

  for (i = 1; i <= 3; i++)
    {
      if (MinSize[2 * i - 1] < MaxSize[2 * i - 1])
	{
	  printf ("DistribPoints: Bad zone sizes.\n");
	  abort ();
	}
      if (MinSize[2 * i] > MaxSize[2 * i])
	{
	  printf ("DistribPoints: Bad zone sizes.\n");
	  abort ();
	}
    }

  id = 0;
  while (id != Qty)
    {
      for (i = 1; i <= 3; i++)
	coo[i] =
	  MaxSize[2 * i - 1] + gsl_rng_uniform (r) * (MaxSize[2 * i] -
						      MaxSize[2 * i - 1]);

      if (coo[1] < MinSize[1] || coo[1] > MinSize[2]
	  || coo[2] < MinSize[3] || coo[2] > MinSize[4]
	  || coo[3] < MinSize[5] || coo[3] > MinSize[6])
	{
	  id++;
	  for (j = 1; j <= 3; j++)
	    Data[id][j] = coo[j];
	}
    }

  oneDDoubleFree (coo);

  return;
}

void
GetPtNextPos (double **Coo, double *Vel, double dt, int interpolation,
	      double *NextCoo)
{
  int i;

  if (interpolation == 1)
    for (i = 0; i <= 1; i++)
      NextCoo[i] = Coo[0][i] + Vel[i] * dt;
  else if (interpolation == 2)
    for (i = 0; i <= 1; i++)
      NextCoo[i] = Coo[1][i] + 2 * Vel[i] * dt;
  else
    abort ();

  return;
}

int PointAligned (double* p1, double* p2, double* p3, double eps)
{
  int i;
  double angle;
  double * v1 = oneDDoubleDAlloc (3);
  double * v2 = oneDDoubleDAlloc (3);
  
  for (i = 0; i < 3; i++)
    v1[i] = p2[i] - p1[i];

  for (i = 0; i < 3; i++)
    v2[i] = p3[i] - p1[i];
  
  angle = VectorAngle (v1, v2, 0, 0);
  
  if (angle < eps)
    return 1;
  else
    return 0;
}




