/* This file is part of the 'neper' program. */
/* Copyright (C) 2004-2009, Romain Quey. */
/* See the COPYING file in the top-level directory. */

#include"MStat.h"

/* CalcVol calculates, records and prints the volume of the elements. */
void
CalcVol (struct NODES Nodes, struct MESH Mesh3D, double **pEltVol)
{
  int i;
  int n1pos, n2pos, n3pos, n4pos;

  (*pEltVol) = ut_alloc_1d (Mesh3D.EltQty + 1);

  /* for every element */
  for (i = 1; i <= Mesh3D.EltQty; i++)
    {
      /* its nodes are recorded: we record their positions. */
      n1pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][0]];
      n2pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][1]];
      n3pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][2]];
      n4pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][3]];
      /*
         printf("nodes: %d(%d) %d(%d) %d(%d) %d(%d)\n",\
         Mesh3D.EltNodes[i][0],n1pos,\
         Mesh3D.EltNodes[i][1],n2pos,\
         Mesh3D.EltNodes[i][2],n3pos,\
         Mesh3D.EltNodes[i][3],n4pos);
       */

      (*pEltVol)[i] = TetrahedronVolume (Nodes.NodesCoo[n1pos],
					 Nodes.NodesCoo[n2pos],
					 Nodes.NodesCoo[n3pos],
					 Nodes.NodesCoo[n4pos]);
    }

  return;
}

/* CalcRho calculates, records and prints the rho-factor of the elements.
 * the supplied results are the same than those of gmsh.
 */
void
CalcRho (struct NODES Nodes, struct MESH Mesh3D, double **pEltRho)
{
  int i;			/* mute variable */
  int n1pos, n2pos, n3pos, n4pos;	/* mute variables */
  double * EdgeLength = ut_alloc_1d (6);		/* mute          */

  (*pEltRho) = ut_alloc_1d (Mesh3D.EltQty + 1);

  /* for every element */
  for (i = 1; i <= Mesh3D.EltQty; i++)
    {
      /* its nodes are recorded: we record their positions. */
      n1pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][0]];
      n2pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][1]];
      n3pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][2]];
      n4pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][3]];

      /* the various edge lengths are calculated. */
      EdgeLength[1] = threeDDoubleDist (Nodes.NodesCoo[n1pos],
					Nodes.NodesCoo[n3pos], 0);
      EdgeLength[0] = threeDDoubleDist (Nodes.NodesCoo[n1pos],
					Nodes.NodesCoo[n2pos], 0);
      EdgeLength[2] = threeDDoubleDist (Nodes.NodesCoo[n1pos],
					Nodes.NodesCoo[n4pos], 0);
      EdgeLength[3] = threeDDoubleDist (Nodes.NodesCoo[n2pos],
					Nodes.NodesCoo[n3pos], 0);
      EdgeLength[4] = threeDDoubleDist (Nodes.NodesCoo[n2pos],
					Nodes.NodesCoo[n4pos], 0);
      EdgeLength[5] = threeDDoubleDist (Nodes.NodesCoo[n3pos],
					Nodes.NodesCoo[n4pos], 0);

      /* rho is calculated */
      (*pEltRho)[i] =
	oneDDoubleMin (EdgeLength, 0, 5) / oneDDoubleMax (EdgeLength, 0, 5);
    }

  oneDDoubleFree (EdgeLength);

  return;
}

void
CalcAngle (struct NODES Nodes, struct MESH Mesh3D, double ***pEltAngle)
{
  int i, j;			/* mute variable */
  double *v1 = ut_alloc_1d (3);
  double *v2 = ut_alloc_1d (3);
  int angleid;
  int face, n1, n2, n3;
  int n1pos, n2pos, n3pos;

  (*pEltAngle) = ut_alloc_2d (Mesh3D.EltQty + 1, 13);

  /* for every element */
  for (i = 1; i <= Mesh3D.EltQty; i++)
    {
      angleid = 0;
      for (face = 1; face <= 4; face++)	/* for every face */
	for (n1 = 1; n1 <= 4; n1++)	/* for every node */
	  {
	    /* skipping n1=face */
	    if (n1 == face)
	      continue;

	    /* searching n2 & n3 */
	    for (n2 = 1; n2 <= 4; n2++)
	      if (n2 != face && n2 != n1)
		break;
	    for (n3 = 1; n3 <= 4; n3++)
	      if (n3 != face && n3 != n1 && n3 != n2)
		break;

	    angleid++;
	    n1pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][n1 - 1]];
	    n2pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][n2 - 1]];
	    n3pos = Nodes.NodesNbInv[Mesh3D.EltNodes[i][n3 - 1]];

	    for (j = 0; j <= 2; j++)
	      {
		v1[j] = Nodes.NodesCoo[n2pos][j] - Nodes.NodesCoo[n1pos][j];
		v2[j] = Nodes.NodesCoo[n3pos][j] - Nodes.NodesCoo[n1pos][j];
	      }
	    (*pEltAngle)[i][angleid] = VectorAngle (v1, v2, 0, 1);
	  }

      (*pEltAngle)[i][0] = oneDDoubleMin ((*pEltAngle)[i], 1, 12);	/* min angle */
    }

  oneDDoubleFree (v1);
  oneDDoubleFree (v2);

  return;
}


void
PrintNStat (struct NODES Nodes, struct MESH Mesh2D, struct MESH Mesh3D,
	    struct IN In)
{
  int i, j, k;
  int nodeonfod;
  int nodeid, nodeidinv;
  int nodeidk;
  FILE *file = OpenFile (In.nstat, "w");
  int **NodeVal = ut_alloc_2d_int (Nodes.NodeQty + 1, 1000);

  int **FoDNodes;
  int **FoD = oneDpIntDAlloc (7);

  /* searching FoDNodes --- */
  FoD = ReadFoD (In.fod);
  FoDNodes = SearchFoDNodes (Mesh2D, FoD);
  /* searched FoDNodes --- */

  /* In the case of a 2nd-order mesh, there is 10 nodes per elt,
   * but the node of 2nd order are not considered for the mesh
   * quality.
   */
  for (i = 1; i <= Mesh3D.EltQty; i++)
    for (j = 0; j <= 3; j++)
      {
	nodeid = Mesh3D.EltNodes[i][j];
	nodeidinv = Nodes.NodesNbInv[nodeid];

	for (k = 0; k <= 3; k++)
	  {
	    if (k == j)
	      continue;

	    nodeidk = Mesh3D.EltNodes[i][k];

	    NodeVal[nodeidinv][0]++;
	    NodeVal[nodeidinv][NodeVal[nodeidinv][0]] = nodeidk;
	  }
      }

  for (i = 1; i <= Nodes.NodeQty; i++)
    {
      if (NodeVal[i][0] == 0)	/* i.e. it's a 2nd-order node. */
	continue;

      nodeid = Nodes.NodesNb[i];

      /* Checking if the node is on a fce of the domain */
      gsl_sort_int (NodeVal[i] + 1, 1, NodeVal[i][0]);
      NodeVal[i][0] = oneDIntCompress (NodeVal[i], 1, NodeVal[i][0]);

      nodeonfod = 0;
      for (j = 1; j <= 6; j++)
	if (oneDIntEltPos (FoDNodes[j], 1, FoDNodes[j][0], nodeid, 0) != -1)
	  nodeonfod = 1;

      if (nodeonfod == 1)	/* i.e. it is on a face of the domain */
	continue;		/* (it's not a real 3d node); skipped. */

      fprintf (file, "%d\t%d\n", nodeid, NodeVal[i][0]);
    }

  CloseFile (file, In.nstat, "w");

  twoDIntFree (NodeVal, Nodes.NodeQty + 1);

  return;
}
