/**********************************************************************
 * Orientation Library - routines for orientation manipulation.       *
 * Copyright (C) 2007, 2008, 2009 Romain Quey                         *
 *                                                                    *
 * This program is free software: you can redistribute it and/or      *
 * modify it under the terms of the GNU General Public License as     *
 * published by the Free Software Foundation, either version 3 of the *
 * License, or (at your option) any later version.                    *
 *                                                                    *
 * This program is distributed in the hope that it will be useful,    *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  *
 * General Public License for more details.                           *
 *                                                                    *
 * You should have received a copy of the GNU General Public License  *
 * along with this program.  If not, see                              *
 * <http://www.gnu.org/licenses/>.                                    *
 *                                                                    *
 * project homepage: http://sourceforge.net/projects/orilib           *
 **********************************************************************/
#include"ol_set.h"

void
ol_set_mean (struct OL_SET set, char* crysym, double *qm)
{
  if (set.size > 0)
    ol_set_mean_ref (set, crysym, set.q[0], qm);

  return;
}

void
ol_set_mean_ref (struct OL_SET set, char* crysym, double *qref, double *qm)
{
  unsigned int i, j;
  double qnorm;
  double *qb = ol_q_alloc ();

  ut_board_1d_zero (qm, 4);

  for (i = 0; i < set.size; i++)
  {
    ol_q_q_qdisori_cur (qref, set.q[i], crysym, qb);

    for (j = 0; j < 4; j++)
      qm[j] += qb[j] * set.weight[i];
  }

  qnorm = ut_board_1d_norm (qm, 4);
  ut_board_1d_scale (qm, 4, 1 / qnorm);

  ol_q_q_q (qref, qm, qm);

  ol_q_free (qb);

  return;
}

int
ol_set_mean_iter (struct OL_SET set, char* crysym, double *qref, double *qm)
{
  double *qb = ol_q_alloc ();
  double theta;
  int it, res = 0;

  it = 0;
  ol_q_memcpy (qref, qb);

  do
  {
    ol_set_mean_ref (set, crysym, qb, qm);
    ol_q_q_disori (qb, qm, crysym, &theta);
    ol_q_memcpy (qm, qb);
    it++;
  }
  while (theta > 1e-1 && it <= 100);

  if (it > 100)
    res = -1;

  ol_q_free (qb);

  return res;
}

struct OL_SET
ol_set_alloc (size_t size)
{
  unsigned int i;
  struct OL_SET Set;

  Set.size = size;
  Set.q = ut_alloc_2d (size, 4);
  Set.weight = ut_alloc_1d (size);
  for (i = 0; i < size; i++)
    Set.weight[i] = 1;

  return Set;
}

void
ol_set_free (struct OL_SET Set)
{
  ut_free_2d (Set.q, Set.size);
  ut_free_1d (Set.weight);

  return;
}

int
ol_set_fscanf (FILE * file, struct OL_SET *pSet)
{
  size_t i;
  double *e = ol_e_alloc ();
  unsigned int size;
  int test;

  /* required because size_t cannot be scanf (???) neither
     with %u nor with %zu with -pedantic (ISO C90) */
  /* adding 'z' is suggested on
     http://manpages.ubuntu.com/manpages/jaunty/man3/scanf.3.html */
  test = fscanf (file, "%u", &size);
  (*pSet).size = (size_t) size;

  if (test == -1)
    return 0;

    return 0;

  (*pSet) = ol_set_alloc ((*pSet).size);

  for (i = 0; i < (*pSet).size; i++)
  {
    if (ol_e_fscanf (file, e) != 1)
    {
      ol_set_free (*pSet);
      return 0;
    }

    ol_e_q (e, (*pSet).q[i]);

    if (fscanf (file, "%lf", &((*pSet).weight[i])) != 1)
    {
      ol_set_free (*pSet);
      return 0;
    }
  }

  ol_e_free (e);

  return 1;
}

int
ol_set_fscanf_short (FILE * file, struct OL_SET *pSet)
{
  size_t i;
  int test = 0;
  double *e = ol_e_alloc ();

  (*pSet).size = ut_file_nblines_pointer (file);

  (*pSet) = ol_set_alloc ((*pSet).size);

  for (i = 0; i < (*pSet).size; i++)
  {
    test = ol_e_fscanf (file, e);
    ol_e_q (e, (*pSet).q[i]);
  }

  ol_e_free (e);

  return test;
}

int
ol_set_fscanf_sample (FILE * file, double factor, struct OL_SET *pSet)
{
  struct OL_SET Setb;

  ol_set_fscanf (file, &Setb);
  ol_set_sample (Setb, factor, pSet);
  ol_set_free (Setb);

  return 1;
}

int
ol_set_fscanf_sample_short (FILE * file, double factor, struct OL_SET *pSet)
{
  struct OL_SET Setb;

  ol_set_fscanf_short (file, &Setb);
  ol_set_sample (Setb, factor, pSet);
  ol_set_free (Setb);

  return 1;
}

int
ol_set_fprintf (FILE * file, struct OL_SET Set, char *format)
{
  unsigned int i;
  int test;
  double *e = ol_e_alloc ();
  char *fmt = ut_alloc_1d_char (4 * strlen (format) + 10);

  sprintf (fmt, "%s %s %s %s\n", format, format, format, format);

  test = fprintf (file, "%u\n", (unsigned int) Set.size);

  for (i = 0; i < Set.size; i++)
  {
    ol_q_e (Set.q[i], e);
    test = fprintf (file, fmt, e[0], e[1], e[2], Set.weight[i]);
  }

  ol_e_free (e);
  ut_free_1d_char (fmt);

  return test;
}

int
ol_set_fprintf_short (FILE * file, struct OL_SET Set, char *format)
{
  unsigned int i;
  int test = 0;
  double *e = ol_e_alloc ();

  for (i = 0; i < Set.size; i++)
  {
    ol_q_e (Set.q[i], e);
    test = ol_e_fprintf (file, e, format);
  }

  ol_e_free (e);

  return test;
}

void
ol_set_subset (struct OL_SET Set1, double factor, struct OL_SET *pSet2)
{
  unsigned int qty, id;
  double factorinv;

  if (factor < 1e-6)
    return;

  factorinv = 1 / factor;

  qty = (unsigned int) ((double) Set1.size * factor) + 2;

  (*pSet2).size = 0;
  (*pSet2).q = ut_alloc_2d (qty, 4);
  (*pSet2).weight = ut_alloc_1d (qty);

  id = (int) (factorinv * .5 - .5);
  while (id < Set1.size)
  {
    ol_q_memcpy (Set1.q[id], (*pSet2).q[(*pSet2).size]);
    (*pSet2).weight[(*pSet2).size] = Set1.weight[id];
    (*pSet2).size++;
    id = (int) (factorinv * ((*pSet2).size + 0.5) - 0.5);
  }

  return;
}

void
ol_set_orthotropic (struct OL_SET Set1, struct OL_SET *pSet2)
{
  unsigned int i, j, id;
  double **qsym = ut_alloc_2d (3, 4);

  (*pSet2) = ol_set_alloc (Set1.size * 4);

  id = 0;
  for (i = 0; i < Set1.size; i++)
  {
    ol_q_refsym (Set1.q[i], 1, qsym[0]);
    ol_q_refsym (Set1.q[i], 2, qsym[1]);
    ol_q_refsym (qsym[0], 2, qsym[2]);

    ol_q_memcpy (Set1.q[i], (*pSet2).q[id++]);
    for (j = 0; j < 3; j++)
      ol_q_memcpy (qsym[j], (*pSet2).q[id++]);
  }

  ut_free_2d (qsym, 3);

  return;
}

void
ol_set_aniso (struct OL_SET set, double **evect, double *eval)
{
  int i, j, k;
  int maxid;
  int n;
  double **R = ut_alloc_2d (set.size, 3);

  double **ERR = ut_alloc_2d (3, 3);
  double *ER = ut_alloc_1d (3);
  double **Sb = ut_alloc_2d (3, 3);

  n = set.size;
  for (i = 0; i < n; i++)
    ol_q_R (set.q[i], R[i]);

  /* calculation of the covariant matrix S */

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
    {
      for (k = 0; k < n; k++)
	ERR[i][j] += R[k][i] * R[k][j];

      ERR[i][j] /= (double) n;
    }

  for (i = 0; i < 3; i++)
  {
    for (k = 0; k < n; k++)
      ER[i] += R[k][i];

    ER[i] /= (double) n;
  }

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      Sb[i][j] = ERR[i][j] - ER[i] * ER[j];

  /* calculation of the eigen values and vectors of S */

  ut_mat_eigen (3, Sb, eval, evect);

  /* 114.59... is "(180 / PI) * 2" */
  for (i = 0; i < 3; i++)
    eval[i] = 114.59155902616464175442 * atan (ut_num_sqrt (eval[i]));

  /* the max coo of v0 and v1 are made positive, and v2
   * is (re)calculated as v0 'vect prod' v1 to get a right-handed
   * csys. */

  ut_mat_transpose (3, evect, evect);

  for (i = 0; i < 3; i++)
  {
    maxid = ut_board_1d_absmax_index (evect[i], 3);
    if (evect[i][maxid] < 0)
      ut_board_1d_scale (evect[i], 3, -1);
  }

  ut_vector_vectprod (evect[0], evect[1], evect[2]);

  ut_mat_transpose (3, evect, evect);

  ut_free_2d (R, set.size);

  return;
}

void
ol_q_aniso_theta (double *q, double **evect, double *theta)
{
  int j, k;
  double tmp;
  double *R = ol_R_alloc ();

  ol_q_R (q, R);

  /* Calculation of the thetas */
  for (j = 0; j < 3; j++)
  {
    tmp = 0;
    for (k = 0; k < 3; k++)
      tmp += evect[j][k] * R[k];

    /* 114.59155902616464175442 is (180 / PI) * 2 */
    theta[j] = 114.59155902616464175442 * atan (tmp);
  }

  ol_R_free (R);

  return;
}

void
ol_set_aniso_thetadistrib (struct OL_SET set, double **evect, double step, double **pdistrib, double *distribfirstx, int *distriblength)
{
  int i, j;
  int n = set.size;
  double **theta = ut_alloc_2d (3, n);
  double *etheta = ut_alloc_1d (3);

  /* Calculation of the thetas */
  for (i = 0; i < n; i++)
  {
    ol_q_aniso_theta (set.q[i], evect, etheta);
    for (j = 0; j < 3; j++)
      theta[j][i] = etheta[j];
  }

  /* Allocation and calculation */
  for (j = 0; j < 3; j++)
  {
    ut_board_distribparam (theta[j], n, step,
			   &(distribfirstx[j]), &(distriblength[j]));

    pdistrib[j] = ut_alloc_1d (distriblength[j]);

    ut_board_distrib (theta[j], n, step, distribfirstx[j],
		      distriblength[j], pdistrib[j]);
  }

  ut_free_2d (theta, 3);
  ut_free_1d (etheta);

  return;
}

void
ol_set_sample (struct OL_SET Set1, double factor, struct OL_SET *pSet2)
{
  unsigned int qty, id;
  double factorinv = 1 / factor;

  qty = (unsigned int) ((double) Set1.size * factor) + 10;

  (*pSet2).size = 0;
  (*pSet2).q = ut_alloc_2d (qty, 4);
  (*pSet2).weight = ut_alloc_1d (qty);

  id = (unsigned int) (factorinv * .5 - .5);
  while (id < Set1.size)
  {
    ol_q_memcpy (Set1.q[id], (*pSet2).q[(*pSet2).size++]);
    (*pSet2).weight[(*pSet2).size] = Set1.weight[id];
    id = (unsigned int) (factorinv * ((*pSet2).size + 0.5) - 0.5);
  }

  return;
}

void
ol_set_disori (struct OL_SET set, char* crysym, double *qmean, double *pdisori)
{
  unsigned int i;
  double tmp;

  (*pdisori) = 0;
  for (i = 0; i < set.size; i++)
  {
    ol_q_q_disori (set.q[i], qmean, crysym, &tmp);
    (*pdisori) += tmp;
  }
  (*pdisori) /= set.size;

  return;
}

int
ol_set_filter (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2)
{
  unsigned int i;
  double *qm = ol_q_alloc ();
  double mdis;
  double *dis = ut_alloc_1d (set1.size);

  ol_set_mean (set1, crysym, qm);

  mdis = 0;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_q_disori (qm, set1.q[i], crysym, &dis[i]);
    mdis += dis[i];
  }
  mdis *= (rel / set1.size);

  (*pset2).size = 0;
  for (i = 0; i < set1.size; i++)
    if (dis[i] < mdis)
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size++]);

  ol_q_free (qm);
  ut_free_1d (dis);

  return set1.size - (*pset2).size;
}

int
ol_set_filter_iter (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2)
{
  ol_set_filter (set1, crysym, rel, pset2);

  while (ol_set_filter (*pset2, crysym, rel, pset2) != 0)
    continue;

  return set1.size - (*pset2).size;
}

int
ol_set_filter_aniso (struct OL_SET set1, char* crysym, double rel, struct OL_SET *pset2)
{
  unsigned int i;
  double *qm = ol_q_alloc ();
  double **Rdis = ut_alloc_2d (set1.size, 3);
  double *qdis = ol_q_alloc ();
  double **v = ut_alloc_2d (3, 3);
  double *mdis = ut_alloc_1d (3);
  double *dis = ut_alloc_1d (3);

  ol_set_mean (set1, crysym, qm);

  for (i = 0; i < set1.size; i++)
    ol_q_q_qdisori (qm, set1.q[i], crysym, qdis);

  ol_set_aniso (set1, v, mdis);

  for (i = 0; i < 3; i++)
    mdis[i] *= rel;

  (*pset2).size = 0;
  for (i = 0; i < set1.size; i++)
  {
    ol_q_aniso_theta (qdis, v, dis);
    if (fabs (dis[0]) < mdis[0] && fabs (dis[1]) < mdis[1]
	&& fabs (dis[2]) < mdis[2])
    {
      ol_q_memcpy (set1.q[i], (*pset2).q[(*pset2).size]);
      (*pset2).size++;
    }
  }

  ol_q_free (qm);
  ut_free_2d (Rdis, set1.size);
  ut_free_1d (qdis);
  ut_free_2d (v, 3);
  ut_free_1d (mdis);
  ut_free_1d (dis);

  return set1.size - (*pset2).size;
}

int
ol_set_filter_aniso_iter (struct OL_SET set1, char* crysym, double rel,
			  struct OL_SET *pset2)
{
  ol_set_filter_aniso (set1, crysym, rel, pset2);

  while (ol_set_filter_aniso (*pset2, crysym, rel, pset2) != 0)
    continue;

  return set1.size - (*pset2).size;
}
